# copyright: Huynh Duong Phuong Vi
# www.sketchupforyou.com
module S4U
    def self.get_file(dir_s4u_name,s4u_name,ext=".txt")
        file_name = s4u_name + ext
        file_name_full = File.join dir_s4u_name,file_name
        if File.exist?(file_name_full)
            file= File.new(file_name_full,"r")
            str=file.readlines
            file.close
            return str.collect{|s| s.chomp}
        end
        return false
        rescue=>error
            p error,error.backtrace.join('\n')
            return false
    end
    def self.set_file(dir_s4u_name,s4u_name,list,ext=".txt")
        file_name = s4u_name + ext
        file_name_full = File.join dir_s4u_name,file_name
        file= File.new(file_name_full, mode: 'w:UTF-8')
        str=list.join("\n")
        file.syswrite str
        file.close
        return true
        rescue=>error
            p error,error.backtrace.join('\n')
            return false
    end
	module S4u_pickup
		class Select
			def initialize()
				ext=IS_OSX ? ".pdf" : ".svg"
				cursor_path= File.join(File.dirname(__FILE__),("icon_mouse" + ext))
				@cursor_id = UI.create_cursor(cursor_path,0,0) if cursor_path

				@scale_factor= Sketchup.version.to_i > 17 ? UI.scale_factor : 1
				@snap=20*@scale_factor

				@pickup_mode_list=[Strings["PickUp Edge Face"],Strings["PickUp Connected Edge"],Strings["PickUp Group Component"]]
				@group_mode_list=[Strings["No Group Objects"],Strings["Group Objects"]]
			end
			def activate
				model=Sketchup.active_model
				@ip=Sketchup::InputPoint.new
				self.get_setting
				self.reset
			end
			def onSetCursor
				UI.set_cursor(@cursor_id) if @cursor_id
			end
			def getExtents
				bb = Sketchup.active_model.bounds
				bb.add @ip.position if @ip.valid?
				return bb
			end
			def get_setting
				@pickup_mode=0
				@group_object=0
				@setting = S4U::S4u_pickup::instance_variable_get :@setting
				if @setting
					@pickup_mode=@setting[0].to_i
					@group_object=@setting[1].to_i
				end

			end
			def deactivate(view)
				@setting=[]
				@setting[0]=@pickup_mode
				@setting[1]=@group_object
				S4U::S4u_pickup::instance_variable_set(:@setting,@setting)
			end
			def setstatus(view=nil)
				text=Strings["Mode"] + ": "
				text+=@pickup_mode_list[@pickup_mode] + " | "
				text+=Strings["Group Mode"] + ": "
				text+=@group_mode_list[@group_object] + " | "
				if @pickup_mode<2
					if @drawingelement
						if @drawingelement.is_a?(Sketchup::Edge) && edge=@drawingelement
							if (@shift_press && @pickup_mode==0) || @pickup_mode==1
								text+=Strings["Connected Edge"]
							else
								text+=Strings["Edge"]+ " | "
								text+=Strings["Press Shift to Extend Connected Edge"]
							end
						elsif @drawingelement.is_a?(Sketchup::Curve) && curve=@drawingelement
							if curve.is_a?(Sketchup::ArcCurve)
								text+=Strings["ArcCurve"]
							else
								text+=Strings["Curve"]
							end
						elsif @drawingelement.is_a?(Sketchup::Face) && face=@drawingelement
							if @surface && @surface.length>1 && !@draw_hidden
								text+=Strings["Surface"]
							else
								text+=Strings["Face"]
							end
						else

						end
					end
				else
					if @drawingelement
						text+=Strings["Press TAB to Select Outside Group"] +  " | "
						text+=@drawingelement.definition.name
					end
				end
				Sketchup.status_text=text
				view.invalidate if view
			end
			def draw(view)
				self.draw_over(view)
			end
			def draw_over(view)
				view.drawing_color = "orange"
				view.line_width = 5
				if @pickup_mode<2
					if @drawingelement
						vers=[]
						if @drawingelement.is_a?(Sketchup::Edge) && edge=@drawingelement
							if (@shift_press && @pickup_mode==0) || @pickup_mode==1
								if !@vertices_edge
									vers=find_vertices(edge)
									if vers.length>1
										@drawingelement=edge
										@trans=@ip.transformation
										@vertices_edge=vers
									end
								end
								if @vertices_edge
									eds=@vertices_edge.collect{|v| v.position.transform @trans}
									view.draw_polyline eds
									view.tooltip= "     " + Strings["Connected Edge"]
								end

							else
								vers=edge.vertices.collect{|v| v.position.transform @trans}
								view.draw_lines vers
								view.tooltip= "     " + Strings["Edge"]
							end

						elsif @drawingelement.is_a?(Sketchup::Curve) && curve=@drawingelement
							if curve.is_a?(Sketchup::ArcCurve)
								view.tooltip= "     " +  Strings["ArcCurve"]
							else
								view.tooltip= "     " +  Strings["Curve"]
							end
							vers=curve.vertices.collect{|v| v.position.transform @trans}
							view.draw_polyline vers
						elsif @drawingelement.is_a?(Sketchup::Face) && face=@drawingelement
							if @surface && @surface.length>1 && !@draw_hidden
								eds=@surface.flat_map{|f| f.edges}
								eds=eds.select{|e| !e.soft? && !e.hidden?}
								view.tooltip= "     " + Strings["Surface"]
							else
								eds=face.edges
								view.tooltip= "     " + Strings["Face"]
							end
							eds=eds.collect{|e| e.vertices.collect{|v| v.position.transform @trans}}
							view.draw_lines eds.flatten
						end
					end
				elsif @drawingelement
					view.tooltip= "     " +  @drawingelement.definition.name
					edges=get_boundingbox(@drawingelement)
					view.draw_lines edges.flatten
				end
			end
			def onMouseMove(flags, x, y, view)
				@draw_hidden=Sketchup.active_model.rendering_options['DrawHidden']
				@shift_press= (CONSTRAIN_MODIFIER_MASK&flags==CONSTRAIN_MODIFIER_MASK)
				@ctrl_press= (COPY_MODIFIER_MASK&flags==COPY_MODIFIER_MASK)
				@alt_press= (ALT_MODIFIER_MASK&flags==ALT_MODIFIER_MASK)
				@flags=((flags&1)==1)
				@ip=nil

				@ip = view.inputpoint x, y
				@pick_helper = view.pick_helper
				@pick_helper.do_pick x,y
				if @pickup_mode<2
					@drawingelement=nil
					@surface=[]
					@vertices_edge=nil
					edge=@ip.edge
					curve=nil
					if !edge && @ip.vertex
						edge=@ip.vertex.edges[0]
					end
					if edge && ((@draw_hidden) || (!@draw_hidden && (!edge.hidden? && !edge.soft?)))
						ps=edge.vertices.collect{|p| view.screen_coords(p.position.transform @ip.transformation) }
						ps.each_index{|i| ps[i].z=0}
						p= Geom::Point3d.new(x,y,0).project_to_line(ps)
						if p && Geom::point_in_polygon_2D( p , ps, true )
							@drawingelement=edge
							@trans=@ip.transformation
						else
							@pick_helper.count.times.each{|i|
									if @pick_helper.leaf_at(i)==edge
										@drawingelement=edge
										@trans=@pick_helper.transformation_at(i)
										break
									end
								}
						end
						if @drawingelement && curve=edge.curve && !((@shift_press && @pickup_mode==0) || @pickup_mode==1)
							if curve=edge.curve
								@drawingelement=curve
							end
						end
					elsif @ip.vertex && @ip.vertex.edges.any?{|e| curve=e.curve }
						@drawingelement=curve
						@trans=@ip.transformation

					elsif @ip.face #&& @pickup_mode==0
						ps=@ip.face.outer_loop.vertices.collect{|p| view.screen_coords(p.position.transform @ip.transformation) }
						ps.each_index{|i| ps[i].z=0}
						if Geom.point_in_polygon_2D([x,y,0]  , ps, true )
							@drawingelement=@ip.face
							@trans=@ip.transformation
						else
							@pick_helper.count.times.each{|i|
								if @pick_helper.leaf_at(i)==@ip.face
									@drawingelement=@ip.face
									@trans=@pick_helper.transformation_at(i)
									break
								end
							}
						end
						@surface=find_surface(@ip.face)[1]
					end
				else
					@drawingelement=nil
					if @ip.valid?
						instance_path=@ip.instance_path.select{|e| e.is_a?(Sketchup::Group) || e.is_a?(Sketchup::ComponentInstance)}
						if instance_path.length>0
							@instance_path_root=instance_path.clone
							@instance_path=[]
							if view.model.active_entities!=view.model.entities
								instance_path.reverse.each_index{|i|
									ins=instance_path.reverse[i]
									if ins.definition!=view.model.active_entities.parent
										@instance_path.unshift(ins)
										@drawingelement=@instance_path[-1]
										@trans=IDENTITY.clone
										@instance_path_root.each{|e|
											if e==@drawingelement
												break
											else
												@trans=@trans*e.transformation
											end
										}
										@instance_index=@instance_path.length-1
									else
										break
									end
								}
							else
								@trans=IDENTITY.clone
								(0..instance_path.length-2).each{|j|
									@trans=@trans*instance_path[j].transformation
								}
								@instance_path=instance_path
								@drawingelement=@instance_path[-1]
								@instance_index=@instance_path.length-1
							end

						end
					end
				end

				self.setstatus(view)
			end
			def get_boundingbox(drawingelement)
				bounds=get_bounds(drawingelement)
				bounds=bounds.collect{|e| e.transform @trans}
				edges=[
				[bounds[0],bounds[1]],[bounds[0],bounds[2]],[bounds[0],bounds[4]],
				[bounds[1],bounds[3]],[bounds[1],bounds[5]],
				[bounds[2],bounds[3]],[bounds[2],bounds[6]],
				[bounds[3],bounds[7]],
				[bounds[4],bounds[5]],[bounds[4],bounds[6]],
				[bounds[5],bounds[7]],[bounds[6],bounds[7]]
				]
				return edges
			end
			def onLButtonUp(flags, x, y, view)
				process_select(view)
				reset(view)
			end
			def onReturn(view)
				process_select(view)
				reset(view)
			end
			def normal_face(ver)
				vx=nil;vy=nil;line=nil
				(0..ver.length-2).each{|i|
					vx=ver[i].vector_to ver[i+1]
					line=ver[i],ver[i+1]
					break if vx.length!=0
				}
				if line
					(0..ver.length-1).each{|i|
						v2=ver[i].project_to_line line
						vy=v2.vector_to ver[i]
						break if vy.length!=0
					}
					return (vx.cross vy) if vx && vy
				end
				Sketchup.active_model.select_tool(nil)
			end
			def process_select(view)
				model=view.model
				entities=model.active_entities
				edit_transform=model.axes.transformation
				model.start_operation TOOL,true,false,false
				if @pickup_mode<2
					g=entities.add_group
					if @drawingelement
						if @drawingelement.valid? && @drawingelement.is_a?(Sketchup::Edge)
							if((@shift_press && @pickup_mode==0) || @pickup_mode==1) && @vertices_edge
								add_edges(g.entities,@vertices_edge)
							else
								add_edge(g.entities,@drawingelement)
							end
						elsif  @drawingelement.valid? && @drawingelement.is_a?(Sketchup::Curve)
							add_curve(g.entities,@drawingelement)
						elsif @drawingelement.valid? && @drawingelement.is_a?(Sketchup::Face)
							add_faces(g.entities,@drawingelement)
						end
						g.transformation=@trans
					end
					if @group_object==0 && g.valid?
						g.explode
					else
						g.name=TOOL
					end
				elsif @drawingelement
					if @instance_index>0
						trans=IDENTITY.clone
						(0..@instance_index).each{|i|
							trans=trans*@instance_path[i].transformation
						}
					else
						trans=@instance_path[0].transformation
					end
					entities.add_instance @drawingelement.definition,trans
				end
				model.commit_operation
				rescue=>error
					model.abort_operation
					p error
					p error.backtrace.join('\n')
			end
			def add_edges(entities,vertices_edge)
				ps=vertices_edge.collect{|v| v.position}
				entities.add_curve(ps)
			end
			def add_curve(entities,curve)
				ps=curve.vertices.collect{|v| v.position}
				if @drawingelement.is_a?(Sketchup::ArcCurve)
					entities.add_arc(curve.center,curve.xaxis,curve.normal,curve.radius,curve.start_angle,curve.end_angle,ps.length-1)
				else
					entities.add_curve(ps)
				end
			end
			def add_edge(entities,edge)
				ps=edge.vertices.collect{|v| v.position}
				entities.add_edges(ps)
			end
			def add_faces(entities,face)
				if @surface && @surface[0] && @surface.length>1 && !@draw_hidden
					@surface.each{|f|
						add_face(entities,f)
					}
				else
					add_face(entities,face)
				end
				clean_edge(entities,true)
				weld_curve(entities)
			end
			def add_face(entities,face)
				mesh=face.mesh(7)
				smooth_flags = Geom::PolygonMesh::AUTO_SOFTEN|Geom::PolygonMesh::SMOOTH_SOFT_EDGES
				entities.add_faces_from_mesh(mesh,smooth_flags,face.material,face.back_material)
			end

			def weld_curve(entities)
				curs=[]
				if @surface && @surface[0] && @surface.length>1 && !@draw_hidden
					edges=@surface.flat_map{|f| f.edges}.uniq
					@surface.each{|f| curs+=f.edges.select{|e| e.is_a?(Sketchup::Edge) && e.curve && ((e.curve.edges&edges)==e.curve.edges)}}
				else
					curs+=@surface[0].edges.select{|e| e.is_a?(Sketchup::Edge) && e.curve && (@surface[0].edges&e.curve.edges)==e.curve.edges}
				end
				curs=curs.collect{|e| e.curve}.uniq
				if Sketchup.version.to_f>20
					curs.each{|c| entities.weld entities.add_curve(c.vertices.collect{|v| v.position})}
				else
					g=entities.add_group
					curs.each{|c| g.entities.add_curve(c.vertices.collect{|v| v.position})}
					g.explode
				end
			end
			def find_surface(fa)
				model=Sketchup.active_model
				faces_soft=[[[fa],nil]]
				faces=[fa]
				all_connected=fa.all_connected.grep(Sketchup::Face).select{|f|
				(f.loops.collect{|l| l.edgeuses.collect{|e| e.edge}}.flatten).any?{|e| e.soft?}}
				i=0
				all_connected=all_connected-faces
				length=all_connected.length
				faces1=[fa]
				eds=[]
				while all_connected.length>0 && i<length
					faces2=[]
					eds1=[]
					faces1.each{|f|
						ed=f.loops.collect{|l| l.edgeuses.collect{|e| e.edge}}.flatten.select{|e| e.soft?}.uniq
						ed=ed - eds
						eds1 += ed
						fs= ed.collect{|e| e.faces}
						fs=fs.flatten.uniq
						fs= fs-faces
						if !fs.empty?
							faces_soft<<[fs,f]
							faces2+=fs
						end
					}
					faces2=faces2.flatten.uniq
					faces2= faces2-faces
					break if faces2.empty?
					faces1=faces2.clone
					eds=eds1.clone
					faces+=faces1
					all_connected=all_connected-faces1
					i+=1
				end
				return faces_soft,faces.uniq
			end
			def clean_edge(entities,delete=false,ang=0.15)
				edge_selected=[]
				edge_selected1=[]
				entities.grep(Sketchup::Edge).each{|e|
					if e.faces.length==0 ||
					(e.faces.length==2 &&
					#e.faces[0].material==e.faces[1].material && e.faces[0].back_material==e.faces[1].back_material &&
					#e.faces[0].normal.samedirection?( e.faces[1].normal))
					(e.faces[0].normal.angle_between(e.faces[1].normal) <= 0.00001))
						edge_selected<<e
					elsif e && e.valid? && (e.faces.length==2) &&
					(e.faces[0].normal.angle_between( e.faces[1].normal) <= Math::PI*ang )
						edge_selected1<<e
					end
				}
				if delete
					edge_selected.each{|e| e.erase! if e.valid? }
				else
					edge_selected1+=edge_selected
				end
				(edge_selected1).compact.each{|e|
					(e.smooth=e.soft=true) if  e && e.valid?
				}
			end
			def getMenu(menu, flags, x, y, view)
				menus=[]
				@pickup_mode_list.each_index{|i| menus<< menu.add_item(@pickup_mode_list[i]) {@pickup_mode=i}}
				menus.each{|m| menu.set_validation_proc(m) { MF_UNCHECKED }}
				menu.set_validation_proc(menus[@pickup_mode]){MF_CHECKED }
				menu.add_separator
				menus=[]
				@group_mode_list.each_index{|i| menus<< menu.add_item(@group_mode_list[i]) {@group_object=i}}
				if @pickup_mode<2
					menus.each{|m| menu.set_validation_proc(m) { MF_UNCHECKED }}
					menu.set_validation_proc(menus[@group_object]){MF_CHECKED }
				else
					menus.each{|m| menu.set_validation_proc(m) { MF_GRAYED }}

				end
			end
			def onKeyDown(key, repeat, flags, view)
				if key==CONSTRAIN_MODIFIER_KEY
					@shift_press=true
					self.setstatus(view)
				elsif ((key==VK_CONTROL) && IS_WIN ) || ((key==VK_ALT) && IS_OSX)
					@ctrl_press=true
					self.setstatus(view)
				elsif (key==VK_ALT && IS_WIN )|| ((key==VK_CONTROL) && IS_OSX)
					@alt_press=true
					self.setstatus(view)
				end

			end
			def onKeyUp(key, repeat, flags, view)
				if (key==CONSTRAIN_MODIFIER_KEY)
					view.lock_inference if view.inference_locked?
					@shift_press=nil
					self.setstatus(view)
				elsif (key==9) || ((key==48) && IS_OSX)
					previous_index()
					self.setstatus(view)
				elsif ((key==VK_CONTROL) && IS_WIN ) || ((key==VK_ALT) && IS_OSX)
					@ctrl_press=nil
					self.setstatus(view)
				elsif (key==VK_ALT && IS_WIN )|| ((key==VK_CONTROL) && IS_OSX)
					@alt_press=nil
					self.setstatus(view)
				elsif (key==VK_RIGHT) || (key==VK_UP)
					next_index()
					self.setstatus(view)
				elsif (key==VK_LEFT)|| (key==VK_DOWN)
					previous_index()
					self.setstatus(view)
				end
			end
			def previous_index()
				if @drawingelement &&  @pickup_mode==2
					if @instance_path.length>0
						if @instance_index>0
							@instance_index-=1
						else
							@instance_index=(@instance_path.length-1)
						end
						@drawingelement=@instance_path[@instance_index]
						@trans=IDENTITY.clone
						@instance_path_root.each{|e|
							if e==@drawingelement
								break
							else
								@trans=@trans*e.transformation
							end
						}
					end
				end
			end
			def next_index()
				if @drawingelement &&  @pickup_mode==2
					if @instance_path.length>0
						if @instance_index<(@instance_path.length-1)
							@instance_index+=1
						else
							@instance_index=0
						end
						@drawingelement=@instance_path[@instance_index]
						@trans=IDENTITY.clone
						@instance_path_root.each{|e|
							if e==@drawingelement
								break
							else
								@trans=@trans*e.transformation
							end
						}
					end
				end
			end
			def onCancel(reason, view)
				self.reset(view)
			end
			def reset(view=nil)
				view.lock_inference if view
				@state=0
				@ip.clear
				@drawingelement=nil
				self.setstatus(view)
			end
			def resume(view)
				self.setstatus(view)
			end

			def get_bounds(element)
				a=element.transformation.to_a
				b=nil
				if element.is_a?Sketchup::ComponentInstance
					b=element.definition.bounds
				elsif element.is_a?Sketchup::Group
					b=element.local_bounds
				else
					return []
				end
				return (0..7).collect{|i| b.corner(i).clone.transform! element.transformation } if b
				return []
			end

			def normal_face(face,t=Sketchup.active_model.edit_transform)
				vx=nil;vy=nil;line=nil
				ver=face.outer_loop.vertices.collect{|v| v.position.transform t}
				(0..ver.length-2).each{|i|
					vx=ver[i].vector_to ver[i+1]
					line=ver[i],ver[i+1]
					if vx.length!=0 && line
						(0..ver.length-1).each{|i|
							v2=ver[i].project_to_line line
							vy=v2.vector_to ver[i]
							break if vy.length!=0
						}
						normal=vx.cross vy if vx && vy
						if normal.length!=0
							facenormal=face.normal.transform t
							normal.reverse! if (normal.angle_between facenormal) >= Math::PI*0.5
							return normal
						end
					end
				}
				return face.normal
			end
			def find_vertices(edge)
				draw_hidden=Sketchup.active_model.rendering_options['DrawHidden']
				all_connected= draw_hidden ? edge.all_connected.select{|e| (e.is_a?Sketchup::Edge)} :
											 edge.all_connected.select{|e| (e.is_a?Sketchup::Edge) && !e.soft? && !e.hidden?}
				edges=[edge]
				v=edge.vertices[0]
				vers=[v]
				ed= draw_hidden ? (v.edges-edges) : (v.edges-edges).select{|e|  !e.soft? && !e.hidden?}
				i=0
				while (ed && ed[0] && ed.length==1) && (i< all_connected.length)
					i+=1
					edges<<ed[0]
					v=ed[0].other_vertex(v)
					vers<<v
					ed= draw_hidden ? (v.edges-edges) : (v.edges-edges).select{|e|  !e.soft? && !e.hidden?}
				end
				v=edge.vertices[1]
				vers.unshift(v)
				ed = draw_hidden ? (v.edges-edges) : (v.edges-edges).select{|e|  !e.soft? && !e.hidden?}
				i=0
				while (ed && ed[0] && ed.length==1) && (i < (all_connected.length))
					i+=1
					edges.unshift ed[0]
					v=ed[0].other_vertex(v)
					vers.unshift v
					ed= draw_hidden ? (v.edges-edges) : (v.edges-edges).select{|e|  !e.soft? && !e.hidden?}
				end
				return vers
			end
		end
		def self.main(mode)
			Sketchup.active_model.select_tool Select.new()
		end
	end
end